
# SimpleInstaller
[![Build Status](https://github.com/solrudev/SimpleInstaller/workflows/Publish/badge.svg)](https://github.com/solrudev/SimpleInstaller/actions?query=workflow%3A%22Publish%22)
![Maven Central](https://img.shields.io/maven-central/v/io.github.solrudev/simpleinstaller.svg)

## Contents
* [Overview](#overview)
    + [Gradle](#gradle)
* [Usage](#usage)
    + [Customization options](#customization-options)
        - [ConfirmationStrategy](#confirmationstrategy)
        - [Notification](#notification)
    + [Installation](#installation)
        - [Install permission](#install-permission)
        - [ApkSource](#apksource)
    + [Uninstallation](#uninstallation)
* [Testing](#testing)
* [Sample app](#sample-app)
* [License](#license)

## Overview
SimpleInstaller is an Android library which provides wrapper over Android packages install and
uninstall functionality leveraging Kotlin coroutines.

It supports Android versions starting from API 16. Split packages installation is also supported
(note that this is only available on Android versions starting from API 21).

SimpleInstaller was developed with deferred execution in mind. You can launch an install or
uninstall session when user is not interacting with your app directly, for example, while foreground
service is running and your application was removed from recents. The way it works is that the user
is shown a high-priority notification which launches a standard Android confirmation by clicking on
it, or full-screen intent with confirmation (depends on firmware).

### Gradle
All versions are available
[here](https://s01.oss.sonatype.org/#nexus-search;gav~io.github.solrudev~simpleinstaller~~~).
```kotlin
implementation("io.github.solrudev:simpleinstaller:x.y.z")
```
Replace "x.y.z" with the latest version.

## Usage

### Customization options
Install or uninstall session may be customized with `SessionOptions`. It allows to set notification
data and different strategies for handling user's confirmation.

Default value can be retrieved from `SessionOptions.DEFAULT` static field.

A new instance may be constructed in a following way:

<details open>
  <summary>Kotlin</summary>
  
```kotlin
val sessionOptions = SessionOptions {
	confirmationStrategy = ConfirmationStrategy.DEFERRED
	notificationData = notificationDataInstance
	// It's also possible to use `notification` DSL function.
}
```
</details>

<details>
  <summary>Java</summary>
  
```java
SessionOptions sessionOptions = new SessionOptions.Builder()
	.setConfirmationStrategy(ConfirmationStrategy.DEFERRED)
	.setNotificationData(notificationDataInstance)
	.build();
```
</details>

#### ConfirmationStrategy
A strategy for handling user's confirmation of installation or uninstallation.
Can be `DEFERRED` (used by default) or `IMMEDIATE`.

`DEFERRED` (default) — user will be shown a high-priority notification (full-screen intent) which will
launch confirmation activity.

`IMMEDIATE` — user will be prompted to confirm installation or uninstallation right away. Suitable for
launching session directly from the UI when app is in foreground.

#### Notification
It is possible to provide notification title, text and icon.

Note: any options for notification will be ignored if `ConfirmationStrategy` is set to `IMMEDIATE`.

If title/text is empty (they are by default), default value is used when notification is displayed.
`android.R.drawable.ic_dialog_alert` is used as a default icon.

Default value can be retrieved from `NotificationData.DEFAULT` static field.

A new instance may be constructed in a following way:

<details open>
  <summary>Kotlin</summary>
  
```kotlin
val notificationData = NotificationData {
	title = "Title"
	contentText = "Text"
	icon = R.drawable.icon
}
```
</details>

<details>
  <summary>Java</summary>
  
```java
NotificationData notificationData = new NotificationData.Builder()
	.setTitle("Title")
	.setContentText("Text")
	.setIcon(R.drawable.icon)
	.build();
```
</details>

### Installation
Installation functionality is provided by `PackageInstaller` interface.

SimpleInstaller has out-of-the-box support for `Uri` (URIs must have `file:` or `content:` scheme),
`AssetFileDescriptor` and `File`. It is possible to subclass [ApkSource](#ApkSource) and pass it to
`PackageInstaller`.

<details open>
  <summary>Kotlin</summary>

```kotlin
// for split packages: packageInstaller.installSplitPackage(apk1, apk2, apk3) { ... }
val result = packageInstaller.installPackage(apk) {
	confirmationStrategy = ConfirmationStrategy.DEFERRED
	notification {
		title = "Notification title"
		contentText = "Notification text"
		icon = R.drawable.icon
	}
}
when (result) {
	InstallResult.Success -> println("Install succeeded.")
	is InstallResult.Failure -> println(result.cause)
}
```

To obtain `PackageInstaller` instance in Kotlin one can just treat it as a singleton `object` because
its companion object implements `PackageInstaller` interface. For example:
```kotlin
val result = PackageInstaller.installPackage(apk)
val packageInstallerInstance = PackageInstaller
```
You can get if `PackageInstaller` has an active session through a property:
```kotlin
val hasActiveSession: Boolean
```
Progress updates can be collected from `progress` SharedFlow property:
```kotlin
val progress: SharedFlow<ProgressData>
```
</details>

<details>
  <summary>Java</summary>

```java
// for split packages: packageInstaller.installSplitPackage(apkSourceArray, sessionOptions, callback)
UriApkSource apkSource = new UriApkSource(apkUri);
packageInstaller.installPackage(apkSource, SessionOptions.DEFAULT, new PackageInstaller.Callback() {
	@Override
	public void onSuccess() {}

	@Override
	public void onFailure(@Nullable InstallFailureCause cause) {}

	@Override
	public void onException(@NonNull Throwable exception) {}

	@Override
	public void onCanceled() {}

	@Override
	public void onProgressChanged(@NonNull ProgressData progress) {}
});
```

To obtain an instance of `PackageInstaller` use static `getInstance()` method:
```java
PackageInstaller packageInstaller = PackageInstaller.getInstance();
```
You can get if `PackageInstaller` has an active session through a getter method:
```java
public boolean getHasActiveSession();
```
Also it's possible to cancel current install session:
```java
public void cancel();
```
</details>

#### Install permission
On Oreo and higher `PackageInstaller` sets an install reason `PackageManager.INSTALL_REASON_USER`,
so on first install there should be a prompt from Android to allow installation. But relying on this
is not recommended, because your process may be restarted if user chooses Always allow, so the result
and progress won't be received anymore. There's `InstallPermissionContract` in `activityresult`
package which you should use to request user to turn on install from unknown sources for your app.

In your Activity:
```kotlin
val requestInstallPermissionLauncher = registerForActivityResult(InstallPermissionContract()) { booleanResult ->
	if (booleanResult) { /* do something */ }
}
```
...
```kotlin
requestInstallPermissionLauncher.launch(Unit)
```

#### ApkSource
SimpleInstaller provides an abstract `ApkSource` class with the following public interface:
```kotlin
val progress: SharedFlow<ProgressData>
abstract val length: Long
open val file: File
abstract fun openInputStream(): InputStream?
abstract suspend fun getUri(): Uri
open fun clearTempFiles()
```
You can provide your own implementation and pass it to `PackageInstaller`'s `installPackage()` or `installSplitPackage()`.

### Uninstallation
Uninstallation functionality is provided by `PackageUninstaller` interface.

<details open>
  <summary>Kotlin</summary>

```kotlin
val result = packageUninstaller.uninstallPackage(packageName) {
	confirmationStrategy = ConfirmationStrategy.DEFERRED
	notification {
		title = "Notification title"
		contentText = "Notification text"
		icon = R.drawable.icon
	}
}
if (result) {
	println("Uninstall succeeded.")
}
```

To obtain `PackageUninstaller` instance in Kotlin one can just treat it as a singleton `object` because
its companion object implements `PackageUninstaller` interface. For example:
```kotlin
val result = PackageUninstaller.uninstallPackage(packageName)
val packageUninstallerInstance = PackageUninstaller
```
You can get if `PackageUninstaller` has an active session through a property:
```kotlin
val hasActiveSession: Boolean
```
</details>

<details>
  <summary>Java</summary>

```java
packageUninstaller.uninstallPackage(packageName, SessionOptions.DEFAULT, new PackageUninstaller.Callback() {
	@Override
	public void onFinished(boolean success) {}

	@Override
	public void onException(@NonNull Throwable exception) {}

	@Override
	public void onCanceled() {}
});
```

To obtain an instance of `PackageUninstaller` use static `getInstance()` method:
```java
PackageUninstaller packageUninstaller = PackageUninstaller.getInstance();
```
You can get if `PackageUninstaller` has an active session through a getter method:
```java
public boolean getHasActiveSession();
```
Also it's possible to cancel current uninstall session:
```java
public void cancel();
```
</details>

## Testing
`PackageInstaller` and `PackageUninstaller` are interfaces, so you can provide your own fake
implementation for tests. For example, you could create an implementation of `PackageInstaller`
which will always return `InstallResult.Failure` with `InstallFailureCause.Storage` cause:
```kotlin
class FailingPackageInstaller : PackageInstaller {

	override val hasActiveSession = false
	override val progress = MutableSharedFlow<ProgressData>()

	private val result = InstallResult.Failure(
		InstallFailureCause.Storage("Insufficient storage.")
	)

	override suspend fun installSplitPackage(vararg apkFiles: ApkSource, options: SessionOptions) = result

	override fun installSplitPackage(vararg apkFiles: ApkSource, options: SessionOptions, callback: PackageInstaller.Callback) {
		callback.onFailure(result.cause)
	}

	override suspend fun installPackage(apkFile: ApkSource, options: SessionOptions) = result

	override fun installPackage(apkFile: ApkSource, options: SessionOptions, callback: PackageInstaller.Callback) {
		callback.onFailure(result.cause)
	}

	override fun cancel() {}
}
```

## Sample app
There's a simple sample app available. It can install chosen APK file and uninstall an application
selected from the installed apps list. Go
[here](https://github.com/solrudev/SimpleInstaller/tree/master/sampleapp) to see sources.

## License
[![License](https://img.shields.io/badge/License-Apache_2.0-blue.svg)](https://github.com/solrudev/SimpleInstaller/blob/master/LICENSE)
